package PBO::WeaverModule;

use strict;
use warnings;

use File::Path qw(make_path);
use File::Basename;
use File::Spec;
use Text::CSV_PP;

use Data::Dumper;

require Exporter;

our @ISA = qw(Exporter);

our %EXPORT_TAGS = ( ALL => [ qw( 
    @inputFiles 
    @params
    @choices
    %invInstNames
    %instList
    $LOG 
    $mode
    $output_dir
    $ext
    $lineTranslationOffset
		$configurator_set
    &mapParamVal
    $param_pattern
    &printInputFiles 
    &printLogFile 
    &collectCommandLineArguments
    &collectPBOConstructs
    &generateConfigFile
	&generateInfoFile
    &generatePropertiesFile
	&generateParamsFile
    &generateTranslationTable
) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{ALL} } );

our @EXPORT = qw( );

our $VERSION = '0.04';
# VERSION HISTORY
# version 0.04 (2013-02-15) - Quinn Hsu
# - fixed pbo.properties not giving enough information
# - fixed properties file to have some intelligent guessing for each parameter/choice
# version 0.03 (2012-08-12) - Quinn Hsu
# - fixed --output_dir being considered as an instantiated 
#	variable when --inst was turned on
# - added read-only status to param_ils param file

# ------------------------- INTRODUCTION -------------------------

# this is the perl module for the weavers of PbO.  this module
# contains all the functions required to parse a set of PbO 
# enriched sources to build the data structures required of PbO.
#
# there exists a very strict template pattern for the invocation
# of function calls for this module.  the template is as follows:
#
# 1. collectCommandLineArguments()	[which this module provides]
# 1. collectPBOConstructs()			[which this module provides]
# 3. generateSupportFile()			[which is in the language specific perl script]
# 4. generateConfigFile()			[which this module provides]
# 5. transformSource()				[which is in the language specific perl script]
# 6. generateInfoFile()				[which this module provides]
# 7. generatePropertiesFile()		[which this module provides]
# 8. generateTranslationTable() 	[which this module provides]

# ---------------------- GLOBAL SHARED VARS ----------------------

# there is coupling here between the module and the
# weavers that use them.  we assume that one of the
# first things a weaver will do is set this @inputFiles
# array.
our @inputFiles;

# array of all declared ##PARAMS that have not been explicitly 
# instantiated by --inst
our @params = ();

# array containing all collected choices
our @choices = ();

# array containing all the log declarations
our @logs = ();

# inverse list of instances needed for building the config file
our %invInstNames = ();

# holds a variable pointing to an open file where all logged information
# will go.  it is the language specific perl script's responsibility
# to open AND close this file.
our $LOG;

# which mode the weaver should operate in ["std","inst","elim"]
#  std mode: expose all choices and parameters
# inst mode: instanciate those params/choices declared on command line
# elim mode: eliminate those choices declared on command line
our $mode = "std";

# the output directory of all the pbo generated files
# this will include 
# 	* pbo.info
#	* pbo.log
#	* pbo.properties
#	* pbo.opt
#	* pbo.translation
#	* pbo_support.c
#	* pbo_support.h
#	* *.c

our $output_dir = "PBO";

# the offset induced by the weaving process of PbO-Enriched sources
# to their target language specific files.  this variable really
# should be set by the target language specific weaver, thus we place
# a dummy value here.
#
# as of this variable's creation, the offsets are
#	* C		: 3
#	* CPP	: 5
#	* Java	: 0

our $lineTranslationOffset = 0;

our $configurator_set = 0;

# the pbo file extension for which the Weaver module should be looking.
# moving the parsing of the input files and pushing them into the queue 
# to the module.

our $ext = "";

# the regular expression used to capture ##PARAM declarations is a
# global shared variable, but is defined below under the ##PARAM REGEXES
# section

# -------------------- GLOBAL VARS FOR MODULE --------------------

# hash of instanciated parameters/choices
# contains { key = name => value = instantiated value }
our %instList = ();

# hash of collected parameter declarations so far
# contains { key = name => 
#     value = %hash{sourcefile} => first encountered declaration }
my %collectedParams = ();

# hash that translates pbo extensions to their language specific 
# extensions

my %extensions = ( 
	"\.pbo-cpp" => "\.cpp",
	"\.pbo-c"	=> "\.c",
	"\.pbo-j"	=> "\.java"
);

# ------------------------ ##PARAM REGEXES -----------------------

# I have decided that, in order to increase readability of the regular
# expressions, I am going to split up the regular expressions into pieces
# and combine them into the final regular expressions as needed.
#
# When first working with the code, this ##PARAM capture pattern was used
# in two seperate places, and it quickly became a hassle when modifying one
# having to remember where the other pattern was.  It was then my goal to
# pre-compile and store in a variable this pattern so that both places
# could reference the same pattern without having me duplicate it.

# using a lot of .+ because \w\d'"\-\+ was not inclusive enough,
# especially when testing with "Hello World!" (did not like exclamation mark)

my $re_type		 			= qr/[\w]+/;
my $re_bound				= qr/.+/;
my $re_bound_separator		= qr/(?:,|\.\.)/;
my $re_name					= qr/[\w]+/;
my $re_default_value		= qr/.+/;
my $re_if_default			= qr/(?(?==) = \s* ($re_default_value))/x;
my $re_optional_bounds		= qr/(?:\[ ($re_bound) \s* $re_bound_separator \s* ($re_bound) \s* \])?/x;

my $re_double_q				= qr/(?=")(".+?")/;
my $re_single_q				= qr/(?=')('.{1}')/;
# everything that's not single or double quoted
#
# strings can have more flexible character classes since
# we really need to capture anything.  but if the string
# to be matched isn't double or single quoted, then chances
# are that the string is a float or integer, and so we MUST
# use a more restrictive character class.  
#
# NOTE: this only relates to categorical string captures
my $re_cat_else				= qr/([\d\.\-\+]+)/;

my $re_cat_pattern			= qr/(?: (?| $re_double_q | $re_single_q | $re_cat_else) ,? )/x;

my $re_opt_range_cat_pattern= qr/(?| $re_optional_bounds | \{(.+)\} )?/x;

# this regex is a bit too permissive in the types of PARAMs that
# it allows to have the optional ranges
#
# at this point in time, we only support ranges for ints and floats
# therefore, when using this pattern to capture information
# we test to see whether the type (\1) is int|integer|float.
#
# in order to force alternation of either ranges or categorical values,
# but not both, i've had to reuse the backreference \2.  in cases where
# a range is found, both \2 and \3 will contain values.  in cases where
# categorical values are found, I further parse them later and this capture
# only happens in \2, leaving \3 undefined/blank.
#
# also note that the categorical regular expression is also too permissive.
# it is entirely possible to mix and match different types within the curly
# braces (i.e. {'a', "abc", 4}) and we will still successfully match against 
# it.  now that I think about it, it is entirely possible to only match against
# valid input, since this regex capturing is done in two steps.  if this issue
# comes up, we may fix it then.

our $param_pattern = qr/\#\#PARAM \s* \( \s* ($re_type) \s* $re_opt_range_cat_pattern \s+ ($re_name) \s* $re_if_default \) /x;
# $1 = type
# $2 = left bound / comma seperated categorical values
# $3 = right bound
# $4 = name
# $5 = default

my $identifier_pattern = qr/\#\#(?!(?:BEGIN|CHOICE|END|LOG|PARAM))($re_name)/x;

my @syntaxErrors = ();

sub printInputFiles { print "input files: @inputFiles\n"; }

sub printLogFile { print $LOG shift };

sub collectCommandLineArguments {

    for my $arg (@_) {
        last if ($arg =~ /^--$/);
        $arg =~ /^--inst$/ && ($mode = "inst") && next;
        # Fixed, --output_dir was being considered as an instantiated variable
        $arg =~ /^--output_dir=(.+)$/ && ($output_dir=$1) && next;
		$arg =~ /^--paramils/i && ($configurator_set=1) && next;;
		
		if ($arg =~ /$ext$/) {
			unless (-e $arg) {
				printf "ERROR: Specified PBO source $arg does not exist.  Please check input file list.\n";
				exit(-1);
			}
			
			push(@inputFiles, $arg);
		}
		
		if ($arg =~ /^--input_files=(.+)$/) {
			my $input_list = $1;
			unless (-e $input_list) {
				printf "ERROR: Specified input list file $input_list does not exist."; 
				exit(-1);
			}
			
			open FL, "<", $input_list or die "ERROR: Could not open file $input_list for reading: $!\n";
			
			while(my $line = <FL>) {
				unless (-e $line) {
					printf "ERROR: File $line specified in file $input_list does not exist.\n";
					exit(-1);
				}
				
				push(@inputFiles, $line);
			}
		}

        # blindly maps on|true|active to 1
        # and maps off|false|inactive to 0
        #
        # should this be the case if the type to be
        # instantiated was actually a string?
        $arg =~ /^--([\w]+) \s* = \s* (.+)/x && ($instList{$1} = &mapParamVal($2));
    }

	if (@inputFiles<1) { 
		printf "ERROR: No $ext files detected.  Please check input files list.\n"; 
		exit(-1); 
	}
	
    printInstList();
    
    if (!-d $output_dir) {
    	printf "Output directory $output_dir does not exist.  Creating directory $output_dir\n";
    	make_path($output_dir);
    }
    
}

sub printInstList() {
    if ($mode eq "inst") {
        printf("\nInstantiating the following parameters and choices:\n");
        
        foreach my $p (keys %instList) {
            printf("  $p=$instList{$p}\n");
        }

    }
}

sub printParams {
# diagnostic: print all param declarations (use after collecting)
	print ("\nparameters found:\n");
	
	for my $i ( 0 .. $#params ) {
		print "$i: ";
		
		for my $field ( keys %{ $params[$i] } ) {
	 		print "$field=$params[$i]{$field} ";
		}
		
		print "\n";
	}
}

sub mapParamVal {
    my($val) = @_;

    if($val eq "on" || $val eq "true" || $val eq "active") {
        $val = "1";
    } elsif ($val eq "off" || $val eq "false" || $val eq "inactive") {
        $val = "0";
    }
    return $val;
}

# collects all ##PARAM declarations over all @inputFiles

sub collectParamDecl {
	# assume that @inputFiles already contains the names of
	# the files to parse
	
	foreach my $sourcename (@inputFiles) {
		open SRC, "<$sourcename" or die ("Unable to open file $sourcename for reading\n");
		
		print $LOG "Collecting parameter definitions from $sourcename ...\n";
		
		my $k = 0;
		
		my $commentDepth = 0;
		
		while(<SRC> ) {
			$k++;
			
			# FIXED
			# We now respect comments of the form //, /* */, /** */
			# However, we need a good language agnostic way of doing this.
			# Since we only support C/C++/Java at the moment, this is fine.
			# But if we go on to support any other languages, a lot of lines in here
			# will need to be changed.
			next if ($_ =~ /\/\/.*?$param_pattern/);
			next if ($_ =~ /\\*.*?$param_pattern.*?\*\//);
			if ($_ =~ /\/\*\*?/) {
				$commentDepth = $commentDepth+1;
			}
			if ($_ =~ /\*\//) {
				$commentDepth = $commentDepth-1;
			}
			
			if ($commentDepth > 0) { next; }
			
			# $1 = type
			# $2 = left bound
			# $3 = right bound
			# $4 = name
			# $5 = default
			
			if ($_ =~ $param_pattern) {
				# Generates warning and then exits
				
				if (exists $collectedParams{$4}) {
					#print "Duplicate param ##$4 found in $sourcename.  First saw ##$4 in $collectedParams{$4}{filename}\n";
					#exit(-1);
					
					push(@syntaxErrors,"ERROR: Duplicate param ##$4 found in $sourcename on line $k.  First saw ##$4 in $collectedParams{$4}{filename} on line $collectedParams{$4}{line}.");
				} else {
					# store as param name as key with the source file as value
					$collectedParams{$4}{filename} = $sourcename;
					$collectedParams{$4}{line} = $k;
					
					my(%p) = ();
					$p{name} = $4;
					$p{type} = $1;
					$p{filename} = $sourcename;
					$p{definition} = $k;
					if (defined $5) {
						# readded mapParamVal because weaving was broken for booleans without
						$p{value} = &mapParamVal($5); 
						$p{default} = 1;
					} else {
					    #revert the change to detect and notify user of undeclared parameter 
					    #push(@syntaxErrors, "ERROR: parameter ##$4 found in $sourcename on line $k must have a default value.");
					    $p{default}=0; 
					}
					
					# this fix below to make sure that ranges are only captured
					# for ints, integers, and floats is a bit of a hack
					#
					# this is because, at the moment, perl does not allow 
					# lookbehind asserts to have variable length.
					#
					# furthermore, i could not figure out a way to get
					# perl to allow me to do a match on a backreference
					# ala \1/int|integer|float/ as part of a condition
					#
					# even more frustrating is that i could not use
					# pattern matching below (i had to use the grep method)
					# because I still need to save the context variable $_
					# to capture these bounds to our $p variable
					#
					# it would have been nice if I could have used
					# defined $2 && defined $3 && $1 =~ /int|integer|float/
					# as the test condition
					#
					# it would have been even nicer if the regex pattern could
					# be rewritten such that:
					#
					# match (type)
					# if (type matches (int|integer|float) match optional range
					# where the conditional match is either (1) a positive
					# lookbehind or (2) a match on the backreference (type)
					
					if (defined $2 && defined $3 && (grep {$_ eq $1} ("int", "integer", "float"))) {
						$p{lbound} = $2;
						$p{rbound} = $3;
					}
					
					# the matches above and below are mutually exclusive.  it
					# cannot be the case that you have both a range for a param
					# and have categorical values.
					#
					# for example,
					#
					# ##PARAM (int[2,3]{4,5,6,7} i)
					#
					# is NOT a legal definition.  furthermore, it is unclear
					# how the regular expression would parse this, even if it 
					# would accept such a string in the first place.
					#
					# since this is not documented as valid syntax, i do not
					# suspect that we will bump up against this case very often.
					
					if (defined $2 && !defined $3) {
						 @{$p{categories}} = $2 =~ /$re_cat_pattern/g;
					}
					
					printf "\nFound param declaration in line $k of $sourcename: \n";
					printf "  name = $4\n";
					printf "  type = $1\n";
					
					if(defined $2 && defined $3 &&(grep {$_ eq $1} ("int", "integer", "float"))) { printf "  bounds: [$2,$3]\n"; }
					if(defined $2 && !defined $3) { print "  categories: {".join(",",@{$p{categories}})."}\n";	}
					if (defined $5) { printf "  default value = $5\n"; }
					else { printf "  no default value\n"; }
					
					printf $LOG "  line $k of $sourcename: $_";
					
					# DeMorgan's Law:
					# !($mode eq "inst") OR !(exists $instList{$2})
					#
					# if $mode is anything other than "inst" we always
					# push the new param into the parameter array
					#
					# if $mode is "inst" then we check to see that
					# this param isn't one of the ones declared for
					# instantiation on the command line
			
					if (!($mode eq "inst" && exists $instList{$4})) {
						push(@params, {%p});
					}
				}
			}
		} #end while
		close(SRC);
	} # end foreach
	
	# we have moved the detection of param declarations to the module file
	# in order to decouple some of the responsibilities of the various functions.
	#
	# in this case, the detection of param declarations was actually in the
	# language specific pbo weaver in the function transformSource.  while this
	# is an artifact of before the creation of the module file, it really doesn't
	# make sense for the transformation stage to still be collecting information.
	#
	# with the change to having all pbo construct mining in this module, we are
	# now also moving &generateInfoFile into the module, where it rightfully belongs
	foreach my $sourcename (@inputFiles) {
		open SRC, "<$sourcename" or die ("ERROR: Could not open $sourcename during collectParamDecl\n");
		print $LOG "Collecting parameter declarations from $sourcename...\n";
		
		my $k=0;
		
		my $commentDepth = 0;
		
		while (my $line = <SRC>) {
			$k++;
			
			# //##x
			if ($line =~ /\/\/.*?$identifier_pattern/) { next; }
			# /* ##x */ or /** ##x */
			if ($line =~ /\/\*\*?.*?$identifier_pattern.*?\*\//) { next; } 
			if ($line =~ /\/\*\*?/) { #/* & /**
				$commentDepth = $commentDepth + 1;
			}
			if ($line =~ /\*\//) {
				$commentDepth = $commentDepth - 1;
			}
			
			if ($commentDepth > 0) { next; }
			
			#if (my @matches = $line =~ /##(?!(?:BEGIN|END|CHOICE|PARAM|LOG))($re_name)/g) {
			#Fixed
			# moved to a pre-compiled regular expression
			
			if (my @matches = $line =~ /$identifier_pattern/g) {
				foreach my $p (@matches) {
					my ($param) = grep {$_->{name} eq $p} @params;
				    #if (!param && !(exists $instList{$p->{name}})) {
					if (!$param) {
						# detect whether we've seen this parameter before
						# this might be because this parameter has been instantiated
						if (!(exists $instList{$p})) {
						
						    # if not, throw an error
						    
						    push(@syntaxErrors, "ERROR: parameter $p, used on line $k of file $sourcename has not been defined.");
						    #print "ERROR: parameter $p, used on line $k of file $sourcename has not been defined.\n";
						    #exit(-1);
						}
					}
					
					if (!defined $param->{declaration} || !defined $param->{declaration}{$sourcename}) {
						# if this is the first time we've seen this param
						# declared, then created the array to hold the line numbers
						$param->{declaration}{$sourcename} = [$k];
					} else {
						push(@{$param->{declaration}{$sourcename}},$k);
					}
				}
			}
		}
		close(SRC);	
	}
	
}

# collect all choice declarations

sub collectChoiceDecl {
	for my $sourcename (@inputFiles) {
		open SRC, "<$sourcename" or die("Unable to open file $sourcename for reading\n");
		
		my $k = 0;
		my @openChoices = ();
		my $skipChoice;
		
		while(<SRC>) {
			$k++;
			
			# // ##BEGIN CHOICE
			# // ##CHOICE
			# // ##END CHOICE
			if ($_ =~ /\/\/.*?\#\#(?:(?:BEGIN|END)\s)?CHOICE/) { next; }
			
			# begin choice declarations
			#
			# $1 name
			# $2 id
			if (/\#\#BEGIN \s+ CHOICE \s+ (\w+) \s* (?(?==)=\s*(\w+))/x) {
				my %choice = ();
				my $name = $1;
			    my $value = $2;

				$choice{name} = $name;
				$choice{startln} = $k;
				$choice{filename} = $sourcename;
				
				my(@conflicts);
				
				if (defined $2) { 
					# we will start to map choice ids to their values here now
					
					$choice{value} = $value; 
					
					@conflicts = grep {$_->{name} eq $name and $_->{value} eq ''} @choices;
					
					if (@conflicts>0) {
					    my $conflict = $conflicts[0];
					    my $conflictName = $conflict->{name};
					    my $conflictFile = $conflict->{filename};
					    my $conflictLine = $conflict->{startln};
					    
					    #print Dumper($conflict);
					    push(@syntaxErrors, "ERROR: choice ##$name with named instance '$value' on line $k of source $sourcename conflicts with choice ##$conflictName on line $conflictLine of source $conflictFile.");
					}
					
					# here we assign the numerical id of the choice instance to 0,1,2,3...
					# which nicely maps to the size of the hash before the add.
					if (!exists $invInstNames{$name} || !exists $invInstNames{$name}{$value}) {
						#print "$name => $value => ". scalar(keys %{$invInstNames{$name}})."\n";
						$invInstNames{$name}{$value} = scalar(keys %{$invInstNames{$name}});
					}
				} else { 
				    @conflicts = grep {$_->{name} eq $name and not $_->{value} eq ''} @choices;
				    if (@conflicts>0) {
				        my $conflict = $conflicts[0];
				        my $conflictName = $conflict->{name};
				        my $conflictValue = $conflict->{value};
				        my $conflictFile = $conflict->{filename};
				        my $conflictLine = $conflict->{startln};
				        
				        push(@syntaxErrors, "ERROR: choice ##$name on line $k of source $sourcename conflicts with choice ##$conflictName with named instance '$conflictValue' on line $conflictLine of source $conflictFile.");
				    }
				    $choice{value} = ""; 
				}
				
				push(@openChoices, {%choice});
				
				print "\nFound beginning of choice $choice{name}".(defined $2 ? "=$choice{value}" : "")." on line $k of source file $sourcename.";
				
				print $LOG "  line $k of $sourcename: $_";
				
				# not too sure about these lines below
				#
				# when merging both begin choice patterns, the lines
				# below were essentially the difference between both
				# blocks of code.
				#
				# i have tried to keep the spirit of these two conditionals
				if ($mode eq "inst" && exists $instList{$1} && defined $2 && $instList{$1} ne $2) {
					$skipChoice=1;
				}
				
				if ($mode eq "inst" && exists $instList{$1} && !defined $2) {
					$skipChoice=1;
				}
				
			}
			
			# choice declarations
			#
			# $1 name
			# $2 id
			if (/\#\#CHOICE \s+ (\w+) \s* = \s* (\w+)/x) {
				#Fixed
				# catch ##CHOICEs that do not have a corresponding ##BEGIN CHOICE
			
				if (@openChoices < 1) {
					push(@syntaxErrors,"ERROR: ##CHOICE $1 does not have an opening ##BEGIN CHOICE $1 on line $k of file $sourcename.");
					next;
				}
			
				my %choice = %{pop @openChoices};
				
				unless ($choice{name} eq $1) {
					#printf "\n\nERROR: ##BEGIN CHOICE $choice{name} on line $choice{startln} does not match ##CHOICE $1 on line $k of file $choice{filename}\n";
					#exit(-1);
					
					push(@syntaxErrors,"ERROR: ##BEGIN CHOICE $choice{name} on line $choice{startln} does not match ##CHOICE $1 on line $k of file $sourcename")
				}
				
				$choice{endln} = $k;
				
				print "\nFound end of choice $1=$choice{value} (started on line $choice{startln}) on line $k of $sourcename.";
				print $LOG "  line $k of $sourcename: $_    (closes choice $1=$choice{value} started on line $choice{startln} of $sourcename.";
				
				if ($mode eq "inst" && exists $instList{$1} && $instList{$1} eq $2) {
					$skipChoice=0;
				} elsif ($mode eq "inst" && exists $instList{$1} && $instList{$1} ne $2) {
					$skipChoice=1;
				} elsif (!$skipChoice) {
					push(@choices, {%choice});
				}
				
				my $name =  $1;
				my $value = $2;
				
				$choice{name}	=$name;
				$choice{value}	=$value;
				$choice{startln}=$k;
				push(@openChoices, {%choice});
				
				if (!exists $invInstNames{$name} || !exists $invInstNames{$name}{$value}) {
					#print "$name => $value => ". scalar(keys %{$invInstNames{$name}})."\n";
					$invInstNames{$name}{$value} = scalar(keys %{$invInstNames{$name}});
				}
				
				print "\nFound beginning of choice $1=$2 on line $k of $sourcename.";
				print $LOG "  line $k: $_";
			
			} elsif (/\#\#CHOICE \s+ (\w+)/x) {
				# it does NOT make sense to have just a 
				# ##CHOICE <name> declaration such as
				#
				# ##BEGIN CHOICE a
				# ##CHOICE b
				# ##END CHOICE a
				#
				# we should throw an error here
				
				push(@syntaxErrors, "ERROR: ##CHOICE $1 on line $k in sourcefile $sourcename must have an id.");
				
				#print "\n\nERROR: ##CHOICE $1 on line $k in sourcefile $sourcename must have an id.";
				#exit(-1);
			}
			
			# end choice declarations
			#
			# $1 name
			
			if (/\#\#END \s+ CHOICE \s+ (\w+)/x) {
				if (@openChoices < 1) {
					push(@syntaxErrors,"ERROR: ##END CHOICE $1 does not have an opening ##BEGIN CHOICE $1 on line $k of file $sourcename");
					next;
				}
			
				my %choice = %{pop @openChoices};
				
				unless ($choice{name} eq $1) {
					#print "\n\nERROR: ##BEGIN CHOICE $choice{name} on line $choice{startln} does not match ##END CHOICE $1 on line $k of $sourcename.\n";
					#exit(-1);
				    
				    push(@syntaxErrors, "ERROR: ##BEGIN CHOICE $choice{name} on line $choice{startln} does not match ##END CHOICE $1 on line $k of $sourcename.");
				}
				
				$choice{endln} = $k;
				
				if ($mode eq "inst" && exists $instList{$1}) {
					$skipChoice=0;
				} elsif (!$skipChoice) {
					push(@choices, {%choice});
				}
				
				print "\nFound end of choice $1".($choice{value} ne "" ? "=$choice{value}" : "")." (started on line $choice{startln}) on line $k of $sourcename.";
				
				print $LOG "  line $k of $sourcename: $_    (closes choice $1".($choice{value} ne "" ? "=$choice{value}" : "") ." started on line $choice{startln}\n";
				
			} # end if	
		} # end while
		
		close(SRC);
		
		if(@openChoices>0) {
			#there are still open choices, we should warn the user
			print "\n\nERROR: there are still ".@openChoices." CHOICEs to be closed\n";
			
			for my $choice (@openChoices) {
			    #print "ERROR: ##CHOICE $choice->{name} in file $choice->{filename} on line $choice->{startln} must be closed\n";
			    push(@syntaxErrors,"ERROR: ##CHOICE $choice->{name} in file $choice->{filename} on line $choice->{startln} must be closed");
			}

            #exit(-1);
		}
		
		print "\n";
	
	}# end for
}

sub collectLogDecl {
	foreach my $sourcefile (@inputFiles) {
		open SRC,"<$sourcefile" or die("ERROR: Could not open file $sourcefile while collecting Log declarations.\n");
		my $k=0;
		while(<SRC>) {
			$k++;
			# $1 is type (for C/C++)
			# $2 is name
			# $3 is value
			if (/\#\#LOG \s* \( ([\w]+) \s+ ([\w]+) \s* = \s* (.+) \) /x) {
				push (@logs, {name=>$2, line=>$k, filename=>$sourcefile});
			}
			
		}
		close(SRC);
	}
	
}

sub collectPBOConstructs {
	&collectParamDecl;
	&collectChoiceDecl;
	&collectLogDecl;
	
	
	&generatePropertiesFile;
	#&generateInfoFile;
	
	if(@syntaxErrors) {
	    for my $error (@syntaxErrors) {
	        print "$error\n";
	    }
	    
	    exit(-1);
	}
}

# NB: it is NOT valid to call generateConfigFile
# without first calling &generateSupportFile in the
# language specific weaver file.
#
# assumes %invInstNames has been filled by 
# &generateSupportFile

sub generateConfigFile {
# generate default configuration file 
#
# (assumes parameter declarations and choices have 
# been collected, instNames have been assigned)

	my $sourcename = join(", ", @inputFiles);
	
	print "Generating pbo.opt ...\n";
	print $LOG "\nGenerating pbo.opt ...\n";
	
	open(DEST, ">$output_dir/pbo.opt");
	
	print DEST <<"END";
; parameter configuration file for $sourcename
; generated by weaver module version $VERSION
; use as follows: main.exe \@pbo.opt
; 
; choices:
END

	my %seenChoiceInst = ();  # hash used for avoiding duplicate inclusion of named instances
	# process once for opt file
	
	my @unique_choices_without_ids = grep { !$seenChoiceInst{$_->{name}}++ } (grep {$_->{value} eq ""} @choices);
	
	foreach my $choice (@unique_choices_without_ids) {
		print DEST "; --$choice->{name}=1\n";
	}
	
	foreach my $choice (keys %invInstNames) {
		foreach my $id (keys %{$invInstNames{$choice}}) {
			print DEST "; --$choice=$invInstNames{$choice}{$id} ; = $id\n";
		}
	}
	
	#for my $i ( 0 .. $#choices ) {
	#	my($name) = $choices[$i]{name};
	#	my($value) = $choices[$i]{value}; 
	#	#printf "$name $value @choiceInst \n";
	#
	#	# if this choice does not have an id
	#	# and I haven't seen it before
	#	if ($value eq "" && !$choiceInst{$name}) { 
	# 		print DEST "; --$name=1\n";
	#  		$choiceInst{$name}=1;
	#
	#		# otherwise, if I haven't seen this choice.id combo
	#		# before
	#	} elsif (!$choiceInst{$name . $value}) {
	# 		print DEST "; --$name=".$invInstNames{$name.'_'.$value}." ; = $value\n";
	#  		$choiceInst{$name . $value}=1;
	#	} 
	#}
	
	print DEST <<"END";
;
; default-less parameters (need to be assigned values):
END

	for my $i ( 0 .. $#params ) {
		if (!$params[$i]{default}) {
			print DEST "; --$params[$i]{name}=\n";
		} 
	}
	
	print DEST <<"END";
; 
; default values:
END

	for my $i ( 0 .. $#params ) {
		if ($params[$i]{default}) {
			# the hack below is required since the opt package will NOT
			# correctly parse the opt file for chars with the format
			# --{name}="{char}" NOR
			# --{name}='{char}'
			# as of opt 3.19
			if (grep {$_ eq $params[$i]{type}} ("char", "character") ) {
				my $value = $params[$i]{value};
				$value =~ s/["']//g;
				print DEST "--$params[$i]{name}=$value\n";
			} else {
	  			print DEST "--$params[$i]{name}=$params[$i]{value}\n";
	  		}
		} 
	}
	
	print DEST <<"END";
; 
; EOF
END
	
	close (DEST);
}

# assumes @choices and @params have been filled in
# by &collectChoiceDecl and &collectParamDecl

sub generatePropertiesFile {
	print "Generating pbo.properties ...\n";
	print $LOG "\nGenerating pbo.properties ...\n";

	open(PROP, ">$output_dir/pbo.properties");
	
	# this hash will hold all the information for choices
	# and params in order to generate the properties file
	my(%properties) = ();
	
	for my $i (0 .. $#choices) {
	    my $choice = $choices[$i];
	    
	    print Dumper($choice);
	
		my($name) = $choices[$i]{name};
		my($value) = $choices[$i]{value};
		my($num_value) = $invInstNames{$value}; # the numerical id of this choice id
	
	
		#print "name: $name value: $value num_value: $num_value\n";
		if (!exists($properties{$name})) {
	  		# create the array if one does not already exist for
	  		# $properties{$name}
	  		
	  		
	  		$properties{$name}{default} = "?";
	  		
	  		if ($value eq "") {
	  		    # choices without named instances can only ever be "on" or "off"
	  		    @{$properties{$name}{values}} = (0,1);
	  		} else {
	  		    # choices with named instances
	  		    @{$properties{$name}{values}} = ();
	  		}
		}
		
		# if this is a choice with named instances, and we haven't seen the named
		# instance before, push it onto the array of valid values.
		if (!($value eq "") and !(grep(/^$value$/, @{$properties{$name}{values}}))) {
		    push(@{$properties{$name}{values}}, $value);
		}
	
		# push onto the array, possibly created above, the
		# numerical value for this choice id
		if (defined $num_value) {
			push(@{$properties{$name}{values}}, "$num_value");
		}
	}
	
	# processing for properties
	for my $i (0 .. $#params) {
		if (!$params[$i]{default}) {
		
		    my ($param) = $params[$i];
		    
		    #print Dumper($param);
		    
	  		my($name) = $params[$i]{name};
	  		# this param neither has a default
	  		# nor values
	  		
	  		@{$properties{$name}{values}} = ("?");
	  		$properties{$name}{default} = "?";
	  		
	  		if (exists $param->{categories}) {
	  		    @{$properties{$name}{values}} = @{$param->{categories}};
	  		} elsif (exists $param->{rbound} and exists $param->{lbound}) {
	  		    @{$properties{$name}{values}}=($param->{lbound}, $param->{rbound});
	  		}
		}
	}
	
	for my $i (0 .. $#params) {
		if ($params[$i]{default}) {
		    my ($param) = $params[$i];
		    
	  		my($name) = $params[$i]{name};
	  		my($value) = $params[$i]{value};
	  
	  		# we may know the default, but we don't
	  		# know the values
	  
	  		$properties{$name}{default} = "$value";
	  		
	  		if (exists $param->{categories}) {
	  		    @{$properties{$name}{values}} = $param->{categories};
	  		} elsif (exists $param->{rbound} and exists $param->{lbound}) {
	  		    push(@{$properties{$name}{values}}, $param->{lbound});
	  		    push(@{$properties{$name}{values}}, $param->{rbound});
	  		} else {
	  		    @{$properties{$name}{values}} = ($param->{value});
	  		}
	  		
	  		
		}
	}
	
	# process properties hash to get all the keys
	my @property_names = ();
	
	
	foreach my $name (keys %properties) {
		push(@property_names, "$name");
	}
	
	# construct a string containing only property names
	# produced from above
	
	print PROP "parameters=".join(",",@property_names)."\n";
	
	foreach my $name (keys %properties) {
		my($value) = join(",",@{$properties{$name}{values}});
		my($default) = $properties{$name}{default};
	
		# if this choice has no ids, then we default it to 
		# have a value of 1
		
		if ($value eq "") { $value = "1"};
		if ($default eq "") { $default = "?"};
		if ($default eq "?" && not ($value eq "" or $value eq "?")) {
		    $default = $properties{$name}{values}->[0];
		} 
	
		print PROP "$name"."_values = $value\n";
		print PROP "$name"."_default = $default\n";
	}
	
	close (PROP);
}

sub generateParamsFile {
	print "Generating pbo_pils_params.txt file ...\n";
	print $LOG "Generating pbo_pils_params.txt file ...\n";
	
	my $file = "$output_dir/pbo_pils_param.txt";
	if (-e $file) { chmod 0644, $file or die ("Error: Could not chmod 0644 $file: $!\n"); }
	
	open DEST, ">", $file or die "ERROR: Could not open file $file for writing: $!\n";
	
	# let's start with choices
	
	my @choices_without_ids = grep {$_->{value} eq ""} @choices;
	
	# for choices without ids, they only map to {0,1}
	
	foreach my $choice (@choices_without_ids) {
		print DEST "$choice->{name} {0,1} [0]\n";
	}
	
	# for choices with instance names, we can reuse $invInstNames since
	# I've already generated all the data
	
	for my $choice (keys %invInstNames) {
		my $ids = join ",", sort(values %{$invInstNames{$choice}});
		print DEST "$choice {$ids} [0]\n";
		
		# generate comments to map between numerical values and the names
		my @nums = sort( values %{ $invInstNames{$choice} } );
		my %r_hash = reverse %{ $invInstNames{$choice} };
		my @mappings = map { "$_=$r_hash{$_}" } @nums;
		
		print DEST "# $choice: ".join(", ", @mappings)."\n";
		
	}
	
	foreach my $param (@params) {
	
	    my $value = (defined $param->{value}) ? $param->{value} : "?";
	
		if (defined $param->{categories}) {
		    # params with categories
			my $ids = join ",", @{$param->{categories}};
			print DEST "$param->{name} {$ids} [$value]\n";
		} elsif ($param->{type} eq "bool" || $param->{type} eq "boolean") {
		    # boolean params
			print DEST "$param->{name} {0,1} [$value]\n";
		} elsif (defined $param->{lbound} && defined $param->{rbound}) {
		    # ints with ranges
			if ($param->{type} eq "int" || $param->{type} eq "integer") {
				# we can enumerate all the values
				my $values = join ",", ($param->{lbound} .. $param->{rbound});
				
				print DEST "$param->{name} {$values} [$value]\n";
			}
		} else {
		    # plain old parameters
		    # no way else to detect these other than not being one of the aboves
		    my $default = ($param->{default}? $param->{value} : "?");
		    
		    print DEST "$param->{name} {?} [$default]\n";
		}
		
	}
	
	chmod 0444, $file or die ("ERROR: Could not set chmod 0444 (readyonly) on file $file: $!\n");
	close(DEST);
	
}

# Debugging function
#
# creates a three-ple as command seperated values that consist of
# 	* the input file name (ending in .pbo-c)
#	* the relative path from above to where the woven .c file got placed
#	* the offset from lines in .pbo-c to the .c file
sub generateTranslationTable {
	my $csv = Text::CSV_PP->new({ eol => "\012", always_quote => 1, escape_char => "\134" });
	my @csv_vals = ();

	for my $file (@inputFiles) {
		# filename without its extension
		my ($filename,$path) = fileparse($file,$ext);
		#my $relPath = File::Spec->abs2rel($file,$output_dir);
		
		my $lang_ext = $extensions{$ext};
		push(@csv_vals, [("$path.$filename$lang_ext", "$path$filename$ext", $lineTranslationOffset)]);
	}
	
	open my $fh, ">:encoding(utf8)", "$output_dir/pbo.translation" or die "ERROR: Could not open pbo.translation for writing: $!\n";
	$csv->print ($fh, $_) for @csv_vals;
	close $fh;
}

# generates information about params, choices, and log statements
# found within the pbo enriched sources.
#
# instead of reworking the entire data structures that get build
# when parsing the sources, we have gone with rebuilding the data
# structures to fit this context.  therefore, each section has two
# for loops, one to rebuild the data structure, and one to go through
# the newly built data structure.

sub generateInfoFile {
	open INFO, ">", "$output_dir/pbo.info" or die "Could not open $output_dir/pbo.info for writing: $!\n";
	
	print INFO <<"END";
// parameters and choices for @inputFiles
// generated by WeaverModule v$VERSION

// parameters
END
	foreach my $param (@params) {
		print INFO "##$param->{name}\n";
		print INFO "  declared: \n    $param->{filename}: $param->{definition}\n";
		print INFO "  accessed:\n";
		foreach my $filename (keys %{$param->{declaration}} ) {
			print INFO "    $filename: @{$param->{declaration}{$filename}}\n";
		}
		
		print INFO "\n";
	}
	
	# the choice array already contains all the information
	# that we would need to construct the data structure to
	# print the information
	
	print INFO <<"END";
	
// choices	
END
	
	my @choicesWithIDs = grep {$_->{value} ne ""} @choices;
	my @choicesWithoutIDs = grep {$_->{value} eq ""} @choices;
	
	# the printout/data structure for choices with ids and without ids
	# will be slightly different, that's why we seperate the two groups
	#
	# however, it is possible that two different choices could
	# have the same name, but one would have ids and the other 
	# would not.  i'm not sure this is a case we should worry about
	
	my %hash = ();
	
	foreach my $choice (@choicesWithoutIDs) {
		my $name = $choice->{name};
		my $filename = $choice->{filename};
		my $lines = {startln=>$choice->{startln},endln=>$choice->{endln}};
		
		if (!defined $hash{$name} || !defined $hash{$name}{$filename}) {
			$hash{$name}{$filename} = [$lines];
		} else {
			push(@{$hash{$name}{$filename}},$lines);
		}
	}
	
	foreach my $choice (keys %hash) {
		print INFO "$choice\n";
		print INFO "  declared:\n";
		foreach my $filename (keys %{$hash{$choice}}) {
			my @lines = map { "$_->{startln}-$_->{endln}" } @{$hash{$choice}{$filename}};
			print INFO "    $filename: @lines\n";
			
			#foreach my $lines (@{$hash{$choice}{$filename}}) {
			#	print "    spans lines: $lines->{startln}-$lines->{endln}\n";
			#}
		}
		
		print INFO "\n";
		
	}
	
	print INFO <<"END";

// choices with instances	
END
	# reuse hash defined above
	
	%hash = ();
	
	foreach my $choice (@choicesWithIDs) {
		my $name = $choice->{name};
		my $value = $choice->{value};
		my $lines = {startln=>$choice->{startln}, endln=>$choice->{endln}};
		my $filename = $choice->{filename};
		
		if (!defined $hash{$name} || !defined $hash{$name}{$value} || !defined $hash{$name}{$value}{$filename}) {
			$hash{$name}{$value}{$filename} = [$lines];
		} else {
			push (@{$hash{$name}{$value}{$filename}},$lines);
		}
	}
	
	foreach my $choice (keys %hash) {
		#print "##$choice\n";
		
		foreach my $id (keys %{$hash{$choice}}) {
			print INFO "$choice=$id\n";
			print INFO "  declared:\n";
			
			foreach my $filename (keys %{$hash{$choice}{$id}}) {
				my @lines = map { "$_->{startln}-$_->{endln}" } @{$hash{$choice}{$id}{$filename}};
				print INFO "    $filename: @lines\n";
				
				#foreach my $lines (@{$hash{$choice}{$id}{$filename}}) {
				#	print "    spans lines: $lines->{startln}-$lines->{endln}\n";
				#}
			}
			print INFO "\n";
		}	
	}
	
	print INFO <<"END";

// logs	
END
	
	%hash=();
	
	foreach my $log (@logs) {
		my $name = $log->{name};
		my $filename = $log->{filename};
		my $line = $log->{line};
		
		if (!defined $hash{$name} || !defined $hash{$name}{$filename}) {
			$hash{$name}{$filename} = [$line];
		} else {
			push (@{$hash{$name}{$filename}}, $line);
		}
	}
	
	foreach my $log (sort keys %hash) {
		print INFO "##$log\n";
		print INFO "  declared:\n";
		
		foreach my $filename (keys %{$hash{$log}}) {
			print INFO "    $filename: @{$hash{$log}{$filename}}\n";
		}
		
		print INFO "\n";
	}
}

1;
