#!/usr/bin/env perl
# pbo-weave-cpp.pl - weave pbo-cpp source file

# call as: pbo-weave-cpp.pl <source_file>
# input: <source_file>
# output: <source_file>.cpp pbo.opt pbo.log 

# (C) Holger H. Hoos, University of British Columbia 2010-12
# version 0.6.06 (2012-09-10) - Quinn Hsu
# - pbo_support.cpp and pbo_support.h no longer get written to the output directory
#   This is so that the makefile generator can work properly
# version 0.6.05 (2012-09-10) - Quinn Hsu
# - fixed a bug where some choices were not being instantiated properly
#   because of the mapParamVal that we do when parsing the command line
# version 0.6.04 (2012-06-05) - Quinn Hsu
# - added the ability for the module files to be in the same
# 		directory as language specific weaver file
# - added #line directive for g++/gdb errors
# - added autotranslation of named choice instances to their
#		integer values
# version 0.6.01 (2012-03-27) - Quinn Hsu
# - removed generateInfoFile (moved to module)
# - fixed problem with choices not being read properly
# version 0.6.00 (2012-02-22) - Quinn Hsu
# - removed all PbO parsing functions to module
# - corrected a regex bug for detecting ##x declarations
# - changed the order command line argument parsing and struct validating
# 		instead, we verify we have all the required command line arguments before
#		we parse them, removing the need to validate the struct for undefined values.

use FindBin;
use lib ($FindBin::Bin);
use strict;

#use warnings;
#use re 'debug';

use PBO::WeaverModule qw(:ALL);
use File::Basename;






my($weaverId) = "weave-pbo-cpp 0.6.06";

$|=1;

$ext = "\.pbo-cpp";

#$mode = "std";  # [std,inst,elim]

# for the C++ weaver, we do not use C++'s string
# objects for the str/string ##PARAMs.  this is 
# to keep both the C and C++ weavers as similar
# as possible.

my %typPC = (
  "bool" => "int", 
  "boolean" => "int", 
  "int" => "long int", 
  "integer" => "long int", 
  "float" => "double",
  "real" => "double",
  "string" => "char*",
  "str" => "char*",
  "character" => "char",
  "char" => "char",
  "double" => "double"
);

my %typPO = (
  "bool" => "OPT_BOOL",
  "boolean" => "OPT_BOOL",
  "int" => "OPT_LONG", 
  "integer" => "OPT_LONG", 
  "float" => "OPT_DOUBLE",
  "real" => "OPT_DOUBLE",
  "string" => "OPT_STRING",
  "str" => "OPT_STRING",
  "character" => "OPT_CHAR",
  "char" => "OPT_CHAR",
  "double"=>"OPT_DOUBLE"
);

my %typUndef = (
  "OPT_BOOL" => "PBO_UNDEF_BOOL",
  "OPT_LONG" => "PBO_UNDEF_INT", 
  "OPT_DOUBLE" => "PBO_UNDEF_FLOAT",
  "OPT_STRING" => "(char*)PBO_UNDEF_STR",  
  "OPT_CHAR" => "PBO_UNDEF_CHR",
);

# TODO: support other param types

#my %paramDeclLine = (); # used for collecting info on param decl
#my %paramAccLines = (); # used for collecting info on param access
#my %choiceLines = (); # used for collecting info on choices
#my %choiceInstNames = (); # used for collecting info on choice instance names
# new in version 0.5:
#my %logLines = (); # used for collecting info on logs

# needed for resolution of ##$LOG statements
my %typPR = (
	"bool" 		=> "\%i", 
	"boolean" 	=> "\%i", 
	"int" 		=> "\%li", 
	"integer" 	=> "\%li", 
	"float" 	=> "\%f",
	"real" 		=> "\%f",
	"character" => "\%c",
	"char" 		=> "\%c",
	"string" 	=> "\%s",
	"str" 		=> "\%s"
);

# end new in version 0.5


sub generateSupportFile {
# modified in version 0.5:
# generate pbo_support.c (assumes that parameter and choice declarations have been collected)
	my $sourcename = join(" ", @inputFiles);

	#print "\n\nGenerating pbo_support.cpp from $sourcename ...\n";
	print $LOG "\nGenerating pbo_support.cpp from $sourcename ...\n";
	
	# first we generate pbo_support.h
	# we will need to add the struct that gets generated
	# within this method to this pbo_support.h file
	# so that this struct can be shared across all
	# pbo-c transformed sources
	
	open(HFILE, ">pbo_support.h");
	print HFILE<<"END";

#ifndef __PBO_SUPPORT__
#define __PBO_SUPPORT__

void pbo_support_init();

void pbo_check_param_values();

int main_before_weaving(int, char**);

int main(int, char**);

END
	
	#open(DEST, ">$output_dir/pbo_support.cpp");
	open(DEST,">","pbo_support.cpp");
	
# you MUST include the extern "C" for MacOS otherwise
# the g++ linker will NOT see the pbo_support struct
	
	print DEST <<"END";
// PbO support in CPP

// (C) Holger H. Hoos, University of British Columbia 2010-11
// version: 0.5 (2011/08/27)

// instantiated from $sourcename by $weaverId


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <string.h>
#include "opt.h"

extern "C" {
#include "pbo_support.h"
}

// constants used to initialise default-less parameters:

static const long int PBO_UNDEF_INT = -9999999;
static const double PBO_UNDEF_FLOAT = -9999999.9;
static const unsigned char PBO_UNDEF_BOOL = 255;
static const void* PBO_UNDEF_STR = NULL;
static const char PBO_UNDEF_CHR = '\\0';

// reference point for cpu timing (= clock state at beginning of execution)
clock_t pbo_clock_start;

// declare the pbo_struct externed in the header file
struct pbo pbo_support;

END

# end modified in version 0.5:

	
	# modified in 0.5.7
	# this struct now gets written to the HFILE we automatically generate
	
	# generate global variables to hold parameters (including those for choices and their values):
	print HFILE "struct pbo {\n";
	print HFILE "// parameters:\n";
	
	for my $i ( 0 .. $#params ) {
		my($tc) = $typPC{$params[$i]{type}};
		if ($tc ne "") {
			print HFILE "  $tc $params[$i]{name};\n";
			
			# new in 0.5.8
			# add ##PARAM range left and right bound variables to the struct
			if (exists $params[$i]{lbound} && exists $params[$i]{rbound}) {
				print HFILE "  $tc $params[$i]{name}"."_lbound;\n";
				print HFILE "  $tc $params[$i]{name}"."_rbound;\n";
			}
			
			# new in 0.5.10
			# add a category array for #PARAMs with categorical values
			if (exists $params[$i]{categories}) {
				
				print HFILE "  $tc $params[$i]{name}"."_categories"."\[".@{$params[$i]{categories}}."\];\n";
			}
			
		} else {  
			print HFILE "\nERROR: Unknown type '$params[$i]{type}'\n";
		}
	}
	
	print HFILE "// choices:\n";
	
	# TODO: break out collection of choice / instance names
	
	# use @choiceNames to collect the names
	# of choices we've seen before
	
	my(@choiceNames) = ();
	for my $i ( 0 .. $#choices ) {
		my($name) = $choices[$i]{name};
	
		# if we haven't seen this choice's name before
		#
		# in 0.5.6, added ^ and $ since this regex was
		# too greedy and would catch on choice names like
		# ##BEGIN CHOICE pay
		# ##BEGIN CHOICE payment
		# if it happen to be the case that pay was processed first
		if (!grep(/^$name$/,@choiceNames)) { # modified in 0.5.6
			print HFILE "  unsigned int $name;\n";
			push (@choiceNames, $name);
		}
	}
	
	# generate constants for named instances of choices (values):
	print HFILE "// constants for choice instance names:\n";
	
	# use instNames to keep track of instance names
	# we've seen before
	my(@instNames) = ();
	for my $i ( 0 .. $#choices ) {
		if ($choices[$i]{value} ne "") {
			my($name) = $choices[$i]{name}.'_'.$choices[$i]{value}; # modified in 0.5.7
			
			if (!grep(/^$name$/,@instNames)) { # modified in 0.5.6
	  			print HFILE "  unsigned int $name;\n";
				push (@instNames, $name);
			}
		}
	}

	print HFILE <<"END";
};

extern struct pbo pbo_support;

#endif

END
	# end modification 0.5.7

	# generate initialisation routine:
	
	print DEST "void pbo_support_init () {\n";
	print DEST "// parameters:\n";
	
	for my $i ( 0 .. $#params ) {
		if ($params[$i]{default}) {
			print DEST "  pbo_support.$params[$i]{name} = $params[$i]{value};\n";
		} else {
			print DEST "  pbo_support.$params[$i]{name} = @typUndef{@typPO{$params[$i]{type}}};\n";
		} 
		# TODO: check types / generate warnings
		
		# new in 0.5.8
		if (exists $params[$i]{lbound} && exists $params[$i]{rbound}) {
			print DEST "  pbo_support.$params[$i]{name}"."_lbound = $params[$i]{lbound};\n";
			print DEST "  pbo_support.$params[$i]{name}"."_rbound = $params[$i]{rbound};\n";
		}
		
		# there is no easy way for us to instanciate the array in the
		# struct other than to loop through each value and assign into
		# the array one value at a time >.<
		if (exists $params[$i]{categories}) {
			for my $j (0 .. @{$params[$i]{categories}}-1) {
				print DEST "  pbo_support.$params[$i]{name}"."_categories[$j]=".$params[$i]{categories}[$j].";\n";
			}
		}
	}
	
	print DEST "// choices (with named instances):\n";
	
	my(%seenChoiceInst) = ();  # hash used for avoiding duplicate inclusion of choices
	
	my @unique_choices_with_ids = grep { !$seenChoiceInst{$_->{name}}++ } (grep {$_->{value} ne ""} @choices);
	foreach my $choice (@unique_choices_with_ids) {
		# we know that if this choice has an id, it must have at least the 0-th id
		# later, we may allow the user to specify which is the default
		print DEST "  pbo_support.$choice->{name} = 0;\n";
	}
	
	#for my $i ( 0 .. $#choices) {
	#	if (!$choiceInst{$choices[$i]{name}} && $choices[$i]{value} ne "") {
	#  		my($ot) = $typPO{"int"};
	#		my($undef) = $typUndef{$ot};
	#		print DEST "  pbo_support.$choices[$i]{name} = $undef;\n";
	#	}
	#}
	
	for my $i ( 0 .. $#choices ) {
		if ($choices[$i]{default}) {
			print DEST "  pbo_support.$params[$i]{name} = $params[$i]{value};\n";
		} 
		# TODO: check types / generate warnings
	}
	
	print DEST "// constants for choice instance names:\n";
	
	#for my $i ( 0 .. $#instNames ) {
	#	my($id) = $i+1;
	#	$invInstNames{$instNames[$i]} = $id;
	#	print DEST "  pbo_support.$instNames[$i]=$id;\n";
	#}
	
	foreach my $choice (keys %invInstNames) {
		foreach my $id (keys %{$invInstNames{$choice}}) {
			#FIXED FOR 0.6.05
			
			#fixed a bug such that choices that were instantiated were still
			# initialized
			if (!exists $instList{$choice}) {
				print DEST "  pbo_support.$choice\_$id=$invInstNames{$choice}{$id};\n";
			}
		}
	}
	
	print DEST "}\n\n";

	# generate routine for checking validity of parameter values (incl. choice parameters):
	#
	# changed in 0.6.00
	# since we can know apriori which parameters don't have defaults (and
	# we must validate over all choices that haven't been forced instantiated)
	# we can validate the command line arguments before passing them to the
	# opt package for parsing.
	# this also has the added benefit of removing the need for magic
	# default uninstantiated values with which we first started out.
	
	print DEST <<"END";
	
void convert_choice_named_instances(int argc, char **argv) {

END
	# changed in 0.6.04
    # assumes uniqueness of named_instances to int pairs
    # which should be guaranteed unless a bug is introduced somewhere
    
    my $num_choices = scalar (keys %invInstNames);
    
    print DEST "// maps for string to int conversion for named instances for choices\n";
      
    foreach my $choice (keys %invInstNames) {
        my @r_hash = sort (values %{ $invInstNames{$choice} });
        my %r_h = reverse %{ $invInstNames{$choice} };
        print DEST  "  char *$choice\_map[".scalar(keys %{$invInstNames{$choice}})."] = {".join(",", map { '"'.$_.'"' }@r_h{@r_hash})."};\n\n";
    }
    
    print DEST "  char *choices[$num_choices] = { ".join(",", map { '"--'.$_.'"' } (keys %invInstNames))." };\n";
    print DEST "  char **choices_map[] = { ".join(", ",map { "$_\_map" } (keys %invInstNames) )." };\n";
    print DEST "  int choices_map_sizes[$num_choices] = { ".join(", ", map { scalar keys %{ $invInstNames{$_} } } (keys %invInstNames) )." };\n";

	print DEST <<"END";
	int i=1,j;
    for(;i<argc;i++) {
        j=0;
        for(;j<$num_choices;j++) {
            char *line = argv[i];
            char *choice = choices[j];
            
            //printf("matching %s with %s\\n", line, choice);
            
            if (strncmp(line,choice,strlen(choice))==0) {
                char *found = strrchr(line,'=');
                
                if (found==NULL) {
                    printf("ERROR: Command line arguments must be of the form: --option=value\\n");
                    exit(-1);
                }
                
                // get the number of named instances of the j'th index of choices_map_sizes
                // get the actual values of this choice from the j'th index of choice_map
                // iterate over all of the values from choice_map, and replace one when we see it
                
                int num_named_instances_for_choice = choices_map_sizes[j];
                char **named_instances = choices_map[j];
                
                int count=0;
                for(;count<num_named_instances_for_choice;count++) {
                    if (strcmp(named_instances[count],found+1)==0) { // found+1 to remove '='
                        
                        char* replace = (char*)malloc(strlen(choice)+1+2+1); 
                        // length of choice + equal sign + the digit for the named instance + '\0'
                        // we assume that you will not have more than 100 named instances for a choice
                        
                        sprintf(replace,"%s=%d",choice,count);
                        
                        //printf("replacement string is: %s\\n", replace);
                        
                        char **addy = &argv[i];
                        *addy=replace;
                        break;
                    }
                } 
                
            }
        }
    }
    
}
	
void pbo_check_param_values (int argc, char** argv) {
		
	int i,j;
END
	
#	for my $i ( 0 .. $#params ) {
#		if ($params[$i]{default}==0) {
#	  		my($ot) = $typPO{$params[$i]{type}};
#			my($undef) = $typUndef{$ot};
#			print DEST <<"END";
#if (pbo_support.$params[$i]{name} == $undef) {
#	printf ("ERROR: Parameter --$params[$i]{name} must be set.\\n");
#	exit(-1);
#}
#END
#		}
#	}

	# find all parameters with no default
	# add "-- and " around the parameter name
	# join these all with a comma
	if (grep {$_->{default}==0} @params) {
		my @defaultless_params = map {'"--'.$_->{name}.'"'} (grep {$_->{default} == 0} @params);
		my $defaultless_params_string = join ",", @defaultless_params;
		
		print DEST <<"END";
	
	int num_defaultless_params = ${\(scalar @defaultless_params)};
	
	char *defaultless_params[${\(scalar @defaultless_params)}] = { $defaultless_params_string };
	
	for(i=0; i<num_defaultless_params; i++) {
		char *source = defaultless_params[i];
		
		int still_scan_arguments = 1;
		
		for(j=1; j<argc; j++) {
			char *line = argv[j];
			char *found = strrchr(line, '=');
			int first_n_chars = found-line;
			if (strncmp(line,source,first_n_chars) == 0) {
				still_scan_arguments = 0;
				break;
			}
		}
		
		if (still_scan_arguments) {
			printf("ERROR: Parameter %s must be set.\\n", source);
			exit(-1);
		}
	}
	
END
	}
	
#	%choiceInst = ();  # hash used for avoiding duplicate inclusion of choices
#	
#	for my $i ( 0 .. $#choices) {
#		if (!$choiceInst{$choices[$i]{name}} && $choices[$i]{value} ne "") {
#	  		my($ot) = $typPO{"int"};
#			my($undef) = $typUndef{$ot};
#			print DEST <<"END";
#if ((int)pbo_support.$choices[$i]{name} == $undef) {
#	printf ("ERROR: Parameter --$choices[$i]{name} must be set.\\n");
#	exit(-1);
#}
#END
#			$choiceInst{$choices[$i]{name}}=1;
#		}
#	}

	if (grep {$_->{value} ne ""} @choices) {

		# first we filter on $choice{value} ne ""
		# then we filter out unique names
		# then we generate an array consisting only of $choice{name}s
		my %seen = ();
		my @unique_choices = map {'"--'.$_->{name}.'"'} (grep {! $seen{$_->{name}}++ } (grep {$_->{value} ne ""} @choices));
		my $unique_choices_string = join ",", @unique_choices;
		
		print DEST <<"END";
	
	int num_unique_choices = ${\(scalar @unique_choices)};
	
	char *unique_choices[${\(scalar @unique_choices)}] = { $unique_choices_string };
	
	for(i=0; i<num_unique_choices; i++) {
		char *source = unique_choices[i];
		
		int still_scan_arguments = 1;
		
		for(j=1; j<argc; j++) {
			char *line = argv[j];
			char *found = strrchr(line, '=');
			int first_n_chars = found-line;
			if (strncmp(line,source,first_n_chars) == 0) {
				still_scan_arguments = 0;
				break;
			}
		}
		
		if (still_scan_arguments) {
			printf("ERROR: Choice %s must be set.\\n", source);
			exit(-1);
		}
	}
	
END
	}
	
	print DEST "}\n\n";

	# added in 0.5.8
	# create a method to check for valid ranges of
	# ##PARAMS that specify ranges
	
	print DEST <<"END";
void pbo_check_param_ranges() {
	
END

	for my $i (0 .. $#params) {
		if (exists $params[$i]{lbound} && exists $params[$i]{rbound}) {
			print DEST <<"END";
  if(!(pbo_support.$params[$i]{name}\_lbound <= pbo_support.$params[$i]{name}) || !(pbo_support.$params[$i]{name} <= pbo_support.$params[$i]{name}\_rbound)) {
	  printf("ERROR: Parameter --$params[$i]{name} must be within the bounds [$params[$i]{lbound}, $params[$i]{rbound}]\\n");
	  exit(-1);
  }
END
		}
	}
	
	print DEST "}\n";	
	
	print DEST <<"END";
void pbo_check_param_categories() {

END

	for my $i (0 .. $#params) {
		if (exists $params[$i]{categories}) {
			if (grep {$params[$i]{type} eq $_} ("str", "string")) {
				my $categories = join(",",@{$params[$i]{categories}});
				$categories =~ s/"/\\"/g;
				
				# we have to escape out the double quotes to make C happy
				# 
				# we do not use ! here (as opposed to the case below for numbers)
				# because strcmp returns 0 on successful match
				print DEST "  if (".join("&&", map {"(strcmp($_, pbo_support.$params[$i]{name}))"} @{$params[$i]{categories}}).") {\n";
				print DEST <<"END";
    printf("ERROR: Parameter --$params[$i]{name} must be one of \{$categories\}\\n");
    exit(-1);
  }\n
END
			} else {
				# must be int, integer, or float
				# for which == is a good comparison operator
				
				# in second print statement, we use a pretend scalar in order
				# to interpolate the function evaluation (join)
				print DEST "  if (".join("&&", map{"($_ != pbo_support.$params[$i]{name})"} @{$params[$i]{categories}}).") {\n";
				print DEST <<"END";
    printf("ERROR: Parameter --$params[$i]{name} must be one of \{${\(join(",",@{$params[$i]{categories}}))}\}\\n");
    exit(-1);
  }\n
END
			}
		}
	}
	
	print DEST "}\n";

	# modified in version 0.5:
	# generate main function: 
	
	print DEST <<"END";
int main_before_weaving(int argc,char *argv[]);

int main(int argc,char *argv[]) {
  pbo_support_init();
  convert_choice_named_instances(argc,argv);
  pbo_check_param_values(argc, argv);
  
END

	for my $i ( 0 .. $#params ) {
		my($ot) = @typPO{$params[$i]{type}};
		
		if ($ot ne "") {
			print DEST "  optrega(&pbo_support.$params[$i]{name}, $ot, '\\0', \"$params[$i]{name}\", \"(PbO)\");\n";
		} else {
			print DEST "ERROR: Unknown type '$params[$i]{type}'\n";
		} 
	}
	
	for my $i ( 0 .. $#choiceNames ) {
		print DEST "  optrega(&pbo_support.$choiceNames[$i], OPT_UINT, '\\0', \"$choiceNames[$i]\", \"(PbO)\");\n";
	}
	
	print DEST <<"END";

  optMain(main_before_weaving);
  optDisableMenu();

  opt(&argc,&argv);
  
  
  pbo_check_param_ranges();
  pbo_check_param_categories();
  
  pbo_clock_start = clock();
  return main_before_weaving(argc,argv);
}

// EOF
END
# end modified in version 0.5

  close(DEST);
  close(HFILE);
}




sub transformSource {
# transform source from $sourcename (assumes parameter declarations have been collected)
	for my $sourcename (@inputFiles) {
		my ($filename,$path) = fileparse($sourcename, $ext);
		
		my $file = "$path.$filename.cpp";
		if (-e $file) {
			unlink($file) or die "WARNING: Could not delete file $file: $!\n";
		}
		
		print "Generating $path.$filename.cpp from $sourcename ...\n";
		print $LOG "\nGenerating $path.$filename.c from $sourcename ...\n";
		
		open(SRC, "<$sourcename");
		open(DEST, ">", $file);
		
		#print DEST "// generated from $sourcename by $weaverId\n\n";

		# you MUST include extern "C" to allow the g++ linker to
		# see the pbo_support struct (for MacOS)

		print DEST <<"END";
extern \"C\" {
#include \"pbo_support.h\"
}
#line 1 \"$filename$ext\"
END
		my($skipChoice)=0;
		my($exitSkipChoice)=0;
		my($k)=0;
		
		while(<SRC>) {
			$k++;
		
			# deal with parameter declarations (##PARAM):
			while($_ =~ $param_pattern) {
				# modified 0.5.10 to capture declaration source name
				
				#$paramDeclLine{$4} = $k;
		
				printf $LOG ("\n  line $k: removed declaration of parameter ##$4\n");
				printf $LOG ("  changed line $k from: $_");
		
				# suppress line
				$_ =~ s/$param_pattern//;
				
				printf $LOG ("  to: $_");
			} 
	
			# deal with choice point declarations (##BEGIN CHOICE, ##CHOICE, ##END CHOICE)
			while (/##BEGIN\s+CHOICE\s+(\w+)\s*=(\w+)/) {
				
				#$choiceLines{$1} = $choiceLines{$1} . "$k";
				#$choiceInstNames{$1} = $choiceInstNames{$1} . "\n  $2 (line $k)";
				
				my($oldln) = $_;
			
				#fixed 0.6.05
				# need to map named choice instances to their mapParamVal equivalents
			
				if ($mode eq "inst" && $instList{$1} eq mapParamVal($2)) {
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)\s*=(\w+)/\/\/ begin choice $1 [instantiated]/;
				} elsif ($mode eq "inst" && exists $instList{$1} && $instList{$1} ne mapParamVal($2)) {
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)\s*=(\w+)/\/\* begin choice $1 [instantiated]/;
					$skipChoice = 1;
				} else {
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)\s*=(\w+)/if (pbo_support.$1==pbo_support.$1_$2) \{ \/\/ begin choice $1/;
				}
			
				printf $LOG ("\n  line $k: resolved ##BEGIN CHOICE $1=$2\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("  to: $_");
			}
		
			while (/##BEGIN\s+CHOICE\s+(\w+)/) {
				
				#$choiceLines{$1} = $choiceLines{$1} . "$k";
				
				my($oldln) = $_;
				
				if ($mode eq "inst" && $instList{$1} eq 1) {  # on
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)/\/\/ begin choice $1 [instantiated]/;
				} elsif ($mode eq "inst" && $instList{$1} eq 0) { # off
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)/\/\* begin choice $1 [instantiated]/;
					$skipChoice = 1;
				} else {
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)/if (pbo_support.$1==1) \{ \/\/ begin choice $1/;
				}
			
				printf $LOG ("\n  line $k: resolved ##BEGIN CHOICE $1\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("  to: $_");
			}
			
			while (/##CHOICE\s+(\w+)\s*=(\w+)/) {
				
				#$choiceInstNames{$1} = $choiceInstNames{$1} . "\n  $2 (line $k)";
				
				my($oldln) = $_;
			
				if ($mode eq "inst" && $instList{$1} eq $2) {
					# note: earlier instances must be stripped
					$_ =~ s/##CHOICE\s+(\w+)\s*=(\w+)/\/\/ choice $1=$2 [instantiated]/;
					$skipChoice = 0;
				} elsif ($mode eq "inst" && exists $instList{$1} && $instList{$1} ne $2) {
					# note: earlier instances may be stripped or not; relaying here on actual stripping of the modified line
					$_ =~ s/##CHOICE\s+(\w+)\s*=(\w+)/\/\/ choice $1=$2 [instantiated]/;
					$skipChoice = 1;
				} else {
					$_ =~ s/##CHOICE\s+(\w+)\s*=(\w+)/\} else if (pbo_support.$1==pbo_support.$1_$2) \{ \/\/ choice $1=$2/;
				}
				
				printf $LOG ("\n  line $k: resolved ##CHOICE $1=$2\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("  to: $_");
			}
			
			while (/##END\s+CHOICE\s+(\w+)/) {
				
				#$choiceLines{$1} = $choiceLines{$1} . "-$k ";
				
				my($oldln) = $_;
					
				if ($mode eq "inst" && exists $instList{$1} && ! $skipChoice) {
					$_ =~ s/##END\s+CHOICE\s+(\w+)/\/\/ end choice $1 [instantiated]/;
				} elsif ($mode eq "inst" && exists $instList{$1} && $skipChoice) {
					$_ =~ s/##END\s+CHOICE\s+(\w+)/  end choice $1 [instantiated] \*\//;
					$exitSkipChoice=1;
				} else {
					$_ =~ s/##END\s+CHOICE\s+(\w+)/\} \/\/ end choice $1/;
				}
				
				printf $LOG ("\n  line $k: resolved ##END CHOICE $1\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("  to: $_");
			}
				
		
			# new in version 0.5:
			# deal with log statements (##$LOG):
			while (/##LOG\s*\(([\w]+)\s+([\w]+)\s*\=\s*(.+)\)/) {
				#$logLines{$2} = $logLines{$2} . "$k ";
			
				printf $LOG ("\n  line $k: resolved log statement (log $2)\n");
				printf $LOG ("  changed line $k from: $_");
			  
				my($time) = "(double)(clock() - pbo_clock_start)/CLOCKS_PER_SEC";
				my($tc) = $typPR{$1};
			
				if ($tc eq "") {
					print DEST "\nERROR: Unknown type '$1'\n";
					$tc = "string";
				}
			  
				my($val) = &mapParamVal($3);  # map boolean constants to standardised values
				$_ =~ s/##LOG\s*\(\s*([\w]+)\s+([\w]+)\s*\=\s*(.+)\)/printf\(\"BEGIN LOG $2 [t=%f]\\n$tc\\nEND LOG $2\\n\", $time, $val\)\;/;
				printf $LOG ("  to: %s", $_);
			} 
			# TODO: check for type consistency
			# end new in version 0.5
		
		
			# deal with parameter references (##<name>):
			while (/##(?!(?:CHOICE|BEGIN|PARAM|END))([\w]+)/) {
				
				# modified in 0.5.10 to include source name
				#$paramAccLines{$1} = $paramAccLines{$1} . "$k ";
				
				my($oldln) = $_;
			
				if ($mode eq "inst" && exists $instList{$1}) {
					$_ =~ s/##([\w]+)/$instList{$1}/;
				} else {
					$_ =~ s/##([\w]+)/pbo_support.$1/;
				}
			
				printf $LOG ("\n  line $k: resolved parameter reference ##$1\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("  to: $_");
			}
		
			# deal with main function (required by opt-3.19)
			if (/main\s*\(.*\)/) {
				printf $LOG ("\n  line $k: renamed original function main ...\n");
				printf $LOG ("  changed line $k from: $_");
				
				$_ =~ s/(int|void)?\s*main\s*\(.*\)/int main_before_weaving(int argc,char *argv[])/; 
				
				printf $LOG ("  to: $_");
				# TODO: catch / handle strange cases (no argc,argv, missing main, ...)
			}
		
			if (! $skipChoice) {
				print DEST $_;
			} else {
				print DEST "\n";
			} 
			
			if ($exitSkipChoice) {
				$skipChoice = 0;
				$exitSkipChoice = 0;
			}
		
		}
		
		close(DEST);
	}
}

# moved generateInfoFile to WeaverModule where it belongs

#sub generateInfoFile {
# generate .info file from parameter and choice information collected from source during transformation
#	my $sourcename = join(", ", @inputFiles);
	
#	open(INFO, ">$output_dir/pbo.info");
#  	print INFO <<"END";
#// parameters and choices in $sourcename
#// generated by $weaverId
#
#// parameters:
#END
#
#	foreach my $key (sort keys %paramAccLines) {
#		# check whether parameter has been declared:
#		
#		if (exists $paramDeclLine{$key}) {
#			printf INFO ("\n##$key:\n  declared in line: $paramDeclLine{$key}\n  used in line(s): $paramAccLines{$key}\n");
#		} else {
#	  		# generate error message:	
#			$paramAccLines{$key} =~ s/\s+$//; # remove trailing white space
#	 		printf ("\nERROR: parameter $key, used in line(s) $paramAccLines{$key}, has not been declared.\n");
#	  		exit(-1);
#		}
#	}
#
#	printf INFO ("\n\n// choices:");
#	
#	foreach my $key (sort keys %choiceLines) {
#		printf INFO ("\n\n##$key:\n  spans lines: $choiceLines{$key}");
#		if ($choiceInstNames{$key} ne "") {
#			print INFO ("\n  includes instances: $choiceInstNames{$key}");
#		}
#	}
#  
# new in version 0.5:
#  	printf INFO ("\n");
#
#TODO: collect and output type of log
# 	printf INFO ("\n\n// logs:\n");
#
#	foreach my $key (sort keys %logLines) {
#		printf INFO ("\n##$key:\n  used in line(s): $logLines{$key}\n");
#	}
# end new in version 0.5
#
# 	print INFO ("\n\n// EOF\n");
#
#  	close(INFO);
#
#}

# Removed in 0.5.7
#
# This has been moved to a seperate utility called pbo-gen-makefile.pl
#
# sub generateMakeFile 


# ------------------ main program ------------------------------------------

#my %p = ();

# step 1: collect all param and choice declarations
# currently assumes only one ##PARAM statement per line (TODO: lift restriction)

&collectCommandLineArguments(@ARGV);

open $LOG, ">", "$output_dir/pbo.log" or die "Could not open $output_dir/pbo.log for writing: $!\n";

printf ("Weaving PbO-CPP files ".join(", ",@inputFiles)." ...\n");
printf $LOG ("Weaving PbO-CPP files ". join(", ",@inputFiles) ." with $weaverId ...\n\n");

&collectPBOConstructs;

#&collectParamDecl;
#&collectChoiceDecl;

# note: log statements are not collected at this point, but just transformed in step 4

# step 2: generate pbo_support.c

&generateSupportFile;

# step 3: generate default configuration file

&generateConfigFile;

# step 4: transform source

&transformSource;

# step 5: generate .info file

&generateInfoFile;
&generatePropertiesFile;
&generateParamsFile;

&generateTranslationTable;

printf ("\n... done.\n");

for my $sourcename (@inputFiles) {
	printf $LOG ("\n... finished weaving $sourcename.\n");
}

close($LOG);

# EOF

exit(0);
